/**
 * @file asio.c - defines implementation for an
 * asynchronous serial IO driver for Propeller C bringup.
 * See header file for more details.
 */
#include "asio.h"
#include <propeller.h>

/*
 * These variables define the receive and transmit queues.
 * The queue pointers will roll over when reaching ASIO_QUEUE_SIZE.
 * ASIO_QUEUE_SIZE will be 2^N - 1 so we can use AND instead of MOD.
 */
int  asio_rxhead;
int  asio_rxtail;
char asio_rxq[ASIO_QUEUE_SIZE+1];

int  asio_txhead;
int  asio_txtail;
char asio_txq[ASIO_QUEUE_SIZE+1];

/*
 * These variables define the interface connection.
 */
int  asio_rx_pinmap;
int  asio_rxflow_pinmap;
int  asio_tx_pinmap;
int  asio_txflow_pinmap;

/*
 * This variable defines the stack used with asio_handler.
 */
long asio_stack[10];

/*
 * This variable defines the tick count required for a bit change.
 */
int  asio_tick;

/**
 * Init interface direction bits and speed.
 * Interface uses "N,1,8": no parity, 1 stop bit, 8 data bits.
 * @note A pin map is a power of 2 integer representing the pin used for a function.
 * If pin 0 is to be the receive pin, then 0x1 (2^0) would be passed via rxmap.
 * @note Flow control parameters are reserved and have no affect today except
 * the direction is set if the pinmap is non-zero for rxflow (backpressure bit).
 * @note The only way to do invert on an interface today is to define override
 * the default constant value. Using a variable would likely change timing.
 * @param rxmap is receive pin map. 
 * @param txmap is transmit pin map. 
 * @param rxflowmap is receive flow control pin map; disabled if 0. 
 * @param txflowmap is transmit flow control pin map; disabled if 0.
 * @param speed is a typedef enum integer indicating interface bit rate. 
 */
void    asio_init (int rxmap, int txmap, int rxflowmap, int txflowmap, ASIO_SPEED_EN speed)
{
    asio_tx_pinmap        = txmap;
    asio_rx_pinmap        = rxmap;
    asio_rxflow_pinmap    = rxflowmap;
    asio_txflow_pinmap    = txflowmap;

    // Init DIRA pinmap in COG. Doing it here won't work.
    
    asio_txhead = 0;
    asio_txtail = 0;
    asio_rxhead = 0;
    asio_rxtail = 0;
    
    asio_tick = (int)speed;
    
   	cognew(asio_handler, asio_stack);
}

/**
 * Show bits on oscilloscope. This is for debug only ... bits come out LSB first
 * @param val is value to show
 * @param mask is the byte/word/long mask
 * @param pinmap is the pin bit map for bit stream output
 * @param syncmap is the bit map for bit stream oscope sync
 * @returns non-zero for looping;
 */
static int show_val (int val, int mask, int pinmap, int syncmap)
{
    int ii = 1;
    // set direction then set word syncmap bit
    asm("or DIRA,%pinmap");
    asm("or DIRA,%syncmap");
    asm("or OUTA,%syncmap"); 
    //val = ~val;
    while(ii & mask) { // this works
        if(val & ii) {
            asm("or   OUTA,%pinmap");
        }
        else {
            asm("andn OUTA,%pinmap");
            asm("andn OUTA,%pinmap"); // waveform symetry
            asm("andn OUTA,%pinmap"); // waveform symetry
        }
        ii = ii << 1;
    }
    asm("andn OUTA,%syncmap");
    return 1;
}

/**
 * Transmit a char on interface.
 * @param val is byte to send
 */
void    asio_tx (char val)
{
    while(asio_txtail != asio_txhead)
        ; // wait for empty queue
    asio_txq[asio_txtail] = val;
    asio_txhead++;
    asio_txhead &= ASIO_QUEUE_SIZE;
}

/**
 * Transmit a string on interface.
 * @param str is string to send.
 */
void    asio_txString (char* str)
{
    while(*str != 0) {
        asio_tx(*str);
        str++;
    }
}

/**
 * Get a byte from the serial interface.
 * param are set by # define constants in asio.h
 * @returns int representing byte received.
 */
int asio_rx (void)
{
    int rc;
    while(asio_rxhead == asio_rxtail) ; // wait for receive char
    //while(!asio_rxready()) ;  // wait for receive char
    rc = asio_rxq[asio_rxtail];
    asio_rxtail++;
    asio_rxtail &= ASIO_QUEUE_SIZE;
    return rc;
}

/*
 * asio_rxready ----> doesn't work ... why not ??? it should <-----
 * Get queue status and return non-zero if byte is available.
 * returns non-zero if rx ready with byte.
int     asio_rxready (void);
int asio_rxready (void)
{
    return (asio_rxhead != asio_rxtail);
}
*/

/**
 * Use calls to asio_rx until a terminator like "\r" is found.
 * All strings returned are null terminated.
 * @param str is a predefined input buffer of len.
 * @param len is maximum number of chars to receive in buffer.
 * @returns number of bytes received.
 */
int asio_rxString (char* str, int len)
{
    int byte = 0;
    int ii = 0;
    do {
        byte = asio_rx();
        str[ii] = byte;
        if(byte == ASIO_ENTER) {
           break;
        }
        ii++;
    } while(ii < len);
    
    if(ii < (len-1)) ii++;
    str[ii] = '\0';
    
    return ii;
}


//#define ASIO_DEBUG 1
/**
 * Handle transmit and receive here.
 * Hard to make our deadline ... again :)
 * LMM/Time lessons learned with beta preview:
 * - Don't ever use a C global with asm("...").
 * - Using globals for decisions takes lots of time. Use locals instead.
 * - Using function calls are expensive (figures) costs 2us for 80MHz prop.
 * - Using "if" in C code is also expensive costs 2us for 80MHz prop.
 * Interestingly, if(--var) ... is cheaper than if(var) ....
 * - Use asm(test ... wz), asm(IF_Z ....) costs 1us for 80MHz prop.
 * - A while(1) costs 0.5us to 0.7us for 80MHz prop.
 * - A for(;;) costs 0.5us to 0.7us for 80MHz prop.
 * - Using too many local variables in a function can throw compiler errors.
 */
void asio_handler(void)
{
    int ii, tmp;
    int time    = 0;
    int count   = ASIO_TICK_COUNT;
   
    int txcnt   = 0;
    int txbyte  = 0;
        
    int rxsync  = 0;
    int rxcnt   = 0;
    int rxbyte  = 0;

    int tick    = asio_tick;
    int ticksav = tick; // much faster than global
    
    // I/O pins must be set on a code-per-cog basis.
    // Setting pins from another cog's code won't work.
    
    tmp = asio_tx_pinmap;
    asm("or     dira, %tmp");       // set outputs
    tmp = asio_rx_pinmap;
    asm("andn   dira, %tmp");       // set inputs
    tmp = asio_rxflow_pinmap;
    asm("or     dira, %tmp");       // rx flow is an output (RTS)
    tmp = asio_txflow_pinmap;
    asm("andn   dira, %tmp");       // tx flow is an input  (CTS)

    tmp = ~ASIO_TX_INVERTED;        // set tx line to idle level
    asm("test       %tmp, #1 wz");  // so first output char is right
    tmp = asio_tx_pinmap;
    asm("IF_Z       andn    OUTA,   %tmp"); // 0? bit = 0;
    asm("IF_NZ      or      OUTA,   %tmp"); // 1? bit = 1;

    /*
     * Basic handler algorithm:
     * Setting/Saving a new rx/tx byte is not time critical;
     * Getting a bit on rx or setting a bit on tx is time critical.
     * Waitcnt in very slow code can cause failure. Current
     * minimum code time is 625 ticks, current tick count 630.
     *
     * Get rx state.
     * If rx state == mark (ASIO_RX_INVERTED) & rx not in progress,
     *    start new byte rx set new wait and continue
     *
     * Else
     *   set tick to bit tick time
     *   If rx byte in progress,
     *      if tick > 0, decrement tick wait and continue
     *      else
     *        set tick to bit tick time
     *        save bit to rxbyte wait and continue
     *   If tx byte in progress,
     *      if tick > 0, decrement tick wait and continue
     *      else
     *        set next bit from txbyte wait and continue
     *   Else if tx tail != head,
     *      load txbyte and tx start bit
     */

#if ASIO_DEBUG
    asm("or      dira, #$c1");
#endif

    for(;;) // forever
    {
        asm("mov        %time, cnt");
        asm("add        %time, %count");

        rxsync = asio_rx_pinmap;
        asm("and        %rxsync, INA wz");
        asm("muxnz      %rxsync, #1");

        rxsync |= rxcnt << 8;        
        // if rxcnt != 0, then tmp != ASIO_RX_INVERTED
        if(rxsync == ASIO_RX_INVERTED) 
        {
#if ASIO_DEBUG
            asm("or     outa, #$80");
#endif
            rxcnt  = 8;     // start another byte only if rxcnt == 0
            rxbyte = 0;     // clear byte rx accumulator
            
            // get a good bit sample point
            tmp  = count >> 0;
            //tmp += count >> 1;     // -600 to +0;
            asm("mov        %time, cnt");
            asm("add        %time, %tmp");
            tick   = ticksav >> 1; // set tick for bit positioning
            // would use for loop, but ICC is out of variables for asm here
            while(tick--) {
                asm("waitcnt    %time, %count");
            }
            tick   = ticksav; // reset tick for new char timing
            
#if ASIO_DEBUG
            asm("andn   outa, #1");
#endif
      
            continue;
        }

        //   if rx byte in progress, get bit in rxbyte
        if(rxcnt > 0)
        {
            // if not tick time, decrement tick and continue
            if(--tick > 0) {
                asm("waitcnt    %time, %count");
                continue;
            }
            // reset tick to bit tick time after tick == 0
            tick = ticksav;
            
#if ASIO_DEBUG
            asm("xor    outa, #1"); // debug only
#endif

            rxsync = ASIO_RX_INVERTED ? ~rxsync : rxsync;
            asm("and    %rxsync, #1");
            rxbyte |= rxsync << (8-rxcnt);
            if(--rxcnt == 0) {    // if rxcnt 0, full byte received. save in buffer.
                if(ASIO_RX_INVERTED) rxbyte ^= 0xff;
                asio_rxq[asio_rxhead] = rxbyte;
                asio_rxhead++;
                asio_rxhead &= ASIO_QUEUE_SIZE;
#if ASIO_DEBUG
                asm("andn outa, #$80");
                asm("andn outa, #1"); // debug only
#endif

                asm("mov        %time, cnt");
                asm("add        %time, %count");    // wait stop-bit time
                for(ii = 0; ii < ticksav; ii++) {
                    asm("waitcnt    %time, %count");
                }
                continue;
            }
            asm("waitcnt    %time, %count");
            continue;
        }
        
        //   if tx byte count in progress, set next bit from txbyte
        if(txcnt > 0)
        {
            // if not tick time, decrement tick and continue
            if(--tick > 0) {
                asm("waitcnt    %time, %count");
                continue;
            }
            // reset tick to bit tick time after tick == 0
            tick = ticksav;
            
            txcnt--;    // only decr txcnt after timer expires
            tmp = asio_tx_pinmap;
            asm("test       %txbyte, #1 wz");
            asm("muxnz      OUTA, %tmp");
            asm("shr        %txbyte, #1");          // txbyte >>= 1;
            asm("waitcnt    %time, %count");
            continue;
        }

        //   else if tx tail != head, load txbyte and tx start bit
        txcnt = 0;
        if(asio_txtail == asio_txhead) {
            asm("waitcnt    %time, %count");
#if ASIO_DEBUG
            asm("andn outa, #$40");
#endif
            continue;
        }

        tmp = asio_txq[asio_txtail];    // get new byte
        txbyte = 0;
        for(ii = 0; ii < 8; ii++) {     // shuffle bits so they go out correctly
            txbyte |= (tmp & (0x80 >> ii)) ? 0 : 1;
            txbyte <<= 1;
        }

        txbyte |= 1; // start bit
        asio_txtail++;
        asio_txtail &= ASIO_QUEUE_SIZE;

        if(!ASIO_TX_INVERTED) txbyte = ~txbyte;
        txcnt = 9 + ASIO_STOP_BITS;

#if ASIO_DEBUG
        asm("or outa, #$40");
#endif

        // end of start tx
        
        // end ... tx
    }
}

//#define ASIO_TEST
#ifdef  ASIO_TEST
int main (int argc, char** argv)
{
    int ii = 0;
    static char rxbuf[ASIO_BUFF_LEN+1]; 

    asio_init(ASIO_PIN_MAP(31), ASIO_PIN_MAP(30), 0, 0, ASIO_57600);
    
    asio_rx(); // block until user opens connection
    asio_txString("Hello World!\r\n");
    
    while(0) {
        asio_tx('U');
    }
    while(0) {
        asio_tx('H');
        asio_tx('i');
        asio_tx(' ');
    }
    while(0) {
        asio_txString("Hello World!\r\n");
    }
    while(0) {
        ii = asio_rx();
        //show_val(ii, 0xff, 0x1, 0x2);
    }
    while(0) {  // echo test
        int ch = asio_rx();
        //for(ii = 0; ii < 500; ii++) ;
        asio_tx(ch);
    }
    while(1) {  // echo string test
        ii = asio_rxString(rxbuf, ASIO_BUFF_LEN);
        //show_val(ii, 0xff, 0x1, 0x2);
        //for(ii = 0; ii < 500; ii++) ; // pause before send
        asio_txString(rxbuf);
    }

    return 0;         
}
#endif
//ASIO_TEST
