/**
 * @file asio.h - Defines interfaces for an
 * asynchronous serial IO driver for Propeller C console.
 * This code works with ICC April Beta release.
 *
 * - Baudrate is defined by supplying constant asio_init.
 *
 * - Flow control is not implemented. Use a 3 wire interface
 * (rx,tx,gnd).
 *
 * - The cognew asio_handler defines 99% of delays required
 * for bit timing using waitcnt assembly commands. This code
 * is tested with protoboard having 5 & 10 MHz crystals.
 * The tx and txString functions work with bauds at
 * 57600, 38400, 28800, 19200, 9600, 4800, 2400, 1200.
 *
 * - It is possible, but not recommended, to try making this
 * module work with Propeller running at less than 80MHz. The
 * asio_handler elements must complete within their 8.68us
 * windows to function; if not the processor will appear to
 * hang waiting for a multi-minute cpu tick count to pass.
 *
 * - Making this code work with > 80MHz propeller will require
 * increasing the value of ASIO_TICK_COUNT for proper bauds.
 *
 * - Using this driver for non-console communications is discouraged.
 *
 * - Receive functions are available, but most likely
 * less useful in their current form as buffering does
 * not seem possible without a way to set up a cog to
 * do the job in the background. A slow baud rate may help.
 * (now that a mechanism is available in beta for new cog,
 * priority will be given to creating a version with queues).
 *
 * - RS232 protocol parameters are {baud,N,1,8} ... stop bits
 * can be set in the hundreds to allow comm with slow
 * devices. This can be configured in the header file.
 *
 * - Public APIs have been tested with propeller pins 0/1
 * connected to a basic stamp module B2SX as well as
 * propeller pins 30/31 with hyperterm and other programs.
 *
 * - Tested on Propeller using 5MHz and 10MHz crystals.
 *
 * - This driver should be used for console communications
 * only as it has trouble with back-to-back characters.
 * If you have trouble with Enter key producing garbage
 * characters on Hyperterm or other terminal, then turn
 * off the option: "Send line ends with line feeds".
 * Often the LF/CR are too close together in time for this
 * driver to differentiate end of one and start of another.
 */
#ifndef __ASIO_H__
#define __ASIO_H__

/**
 * Macro to get pinmap
 */
#define ASIO_PIN_MAP(pin) (1 << pin)

/**
 * Define TX pin map. This code works with ICC April Alpha update.
 * Early ICC March 29 Alpha release has issues with asm(...) parameters.
 * Hard-coding the setbit asm call will make this work. i.e. asm("or  OUTA, $40000000");
 */
#define ASIO_TX_PIN_MAP ASIO_PIN_MAP(30)
/**
 * Define RX pin map. Using bit 31 for rx works fine.
 */
#define ASIO_RX_PIN_MAP ASIO_PIN_MAP(31)

/*
 * Wait tick count for a bit period loop
 */
#define ASIO_TICK_COUNT  630 // 625 min ... use 1010 for 38400 ASIOP(2) ....

/*
 * Speed eunm tick period adjust macro
 */
#define ASIOP(tick) (tick)

/**
 * Typedef enum defining speed to use for the interface.
 * The number in the enum is actually the tick period used by the driver.
 */
typedef enum asio_speed
{
    ASIO_57600    = ASIOP(2),
    ASIO_38400    = ASIOP(3),
    ASIO_28800    = ASIOP(4),
    ASIO_19200    = ASIOP(6),
    ASIO_9600     = ASIOP(12),
    ASIO_4800     = ASIOP(24),
    ASIO_2400     = ASIOP(48),
    ASIO_1200     = ASIOP(96)
}   ASIO_SPEED_EN;

/**
 * Use large number (100+) of TX stop or "idle" bits
 * for slow interfaces adjust at your liesure.
 */
#define ASIO_STOP_BITS      2

/**
 * Use this define to invert the interface sense (or not)
 */
#define ASIO_RT_INVERTED    0   // just 1 or 0 ... anything else breaks code
/**
 * Use this define to invert the TX interface sense (or not)
 */
#define ASIO_TX_INVERTED    ASIO_RT_INVERTED
/**
 * Use this define to invert the RX interface sense (or not)
 */
#define ASIO_RX_INVERTED    ASIO_RT_INVERTED

#define ASIO_ENTER '\r'
#define ASIO_QUEUE_SIZE 0xFF    // queue size and buffer head/tail mask
#define ASIO_BUFF_LEN   ASIO_QUEUE_SIZE     

/*
 * Forward declaration for cog rx/tx handler.
 */
void asio_handler(void);


/*
 * API functions. See implementation for usage.
 */
void    asio_init (int rxmap, int txmap, int rxflowmap, int txflowmap, ASIO_SPEED_EN speed);
int     show_val(int val, int mask, int pinmap, int syncmap);
void    asio_tx(char c);
void    asio_txString(char* str);
int     asio_rx(void);
int     asio_rxString(char* str, int len);

 
#endif
//__ASIO_H__


